// vincia01.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to run Z decays at LEP I
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (C) 2019 Peter Skands

// PYTHIA and VINCIA include files and namespaces
#include "Pythia8/Vincia.h"
using namespace Pythia8;

// ROOT: include files for linking to ROOT, uncomment as necessary
// #include "TROOT.h"
// #include "TDirectory.h"
// #include "TFile.h"
// #include "TH1.h"
// #include "TH2.h"

// Main Program
int main() {
  
  //************************************************************************
  // Define Pythia 8 generator with plug-in
  
  Pythia pythia;
  VinciaPlugin vincia(&pythia);
  
  // Read user settings from file
  pythia.readFile("vincia01.cmnd");

  //************************************************************************
  
  // Shorthands
  Event& event = pythia.event;
  Settings& settings = pythia.settings;
  
  // Extract settings to be used in the main program.
  int    nEvent     = settings.mode("Main:numberOfEvents");
  int    nAbort     = pythia.mode("Main:timesAllowErrors");  
  bool   vinciaOn   = settings.flag("Vincia");
  int    verbose    = settings.mode("Vincia:verbose");
  bool   helicityOn = settings.flag("Vincia:helicityShower");
  int    iEventPri  = -1;
  double eCM        = settings.parm("Beams:eCM");
  
  //************************************************************************
  
  // Initialize
  if(!vincia.init()) { return EXIT_FAILURE; }
  
  //************************************************************************
  
  // Define a few PYTHIA utilities
  Thrust Thr(1);
  Sphericity SphLin(1,2);
  
  //************************************************************************
  
  // Define PYTHIA histograms
  // Rootlink: to use ROOT instead, HIST -> TH1F ("tag",...)
  
  // Number of quarks, partons, and charged particles
  double nQsum      = 0.0;
  double nPsum      = 0.0;
  double nCsum      = 0.0;
  double nStrangeSum = 0.0;
  double nCharmSum  = 0.0;
  double nBottomSum = 0.0;
  double nBaryonSum = 0.0;
  double nGammaSum  = 0.0;
  Hist HnQuarks("nQuarks", 50, -0.5, 49.5);
  Hist HnPartons("nPartons", 100, -0.5, 99.5);
  Hist HnCharged("nCharged", 200, -1.0, 199.0);
  Hist HnGamma("nPhotons", 50, -0.5, 49.5);
  Hist HxStrange("LnXstrange", 10, -5.0, 0.0);
  Hist HxCharm("LnXcharm", 10, -5.0, 0.0);
  Hist HxBottom("LnXbottom", 10, -5.0, 0.0);
  Hist HxFirst("Ln(xFirst)",100,-7.5,0.0);
  Hist HxSecond("Ln(xSecond)",100,-7.5,0.0);
  Hist HxSecondQuark("Ln(xSecondQuark)",100,-7.5,0.0);
  Hist HmSS("mSS",50,0.,10.);
  Hist HmCC("mCC",50,0.,25.);
  Hist HmBB("mBB",50,0.,50.);
  
  // Thrust
  int sBins=100;
  int sBinsCoarse=25;
  Hist Hthr("1-T",sBins, 0.0, 0.5);
  Hist HthrTail("1-T", sBinsCoarse, 0.0, 0.5);
  double wthr=sBins/0.5;
  double wthrCoarse=sBinsCoarse/0.5;
  // Event Shapes
  double wShape=sBins/1.0;
  Hist Hmaj("Major",sBins,0.0,1.0);
  Hist Hmin("Minor",sBins,0.0,1.0);
  Hist Hobl("Oblateness",sBins,0.0,1.0);
  Hist HistC("C",sBins,0.0,1.0);
  Hist HistD("D",sBins,0.0,1.0);
  
  //************************************************************************
  
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  
  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settings.mode("Random:seed");
  vincia.rndmPtr->init(initseed);
  
  
  // Counter for negative-weight events
  double evwt=1.0;
  double sumWeights = 0.0;
  
  // Begin event loop
  int iAbort = 0;
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    bool aborted = !pythia.next();
    aborted = ( aborted || vincia.vinCom.doAbort()) ;
    if(aborted){
      event.list();
      if (++iAbort < nAbort){
	vincia.vinCom.setAbort(false);
	continue;
      }
      cout << " Event generation aborted prematurely, owing to error!\n";
      cout<< "Event number was : "<<iEvent<<endl;
      break;
    }
    
    // Check for weights
    if (vinciaOn) evwt = vincia.weight();
    sumWeights += evwt;
    
    // Print event with helicities
    if (iEvent == 0 || iEvent == iEventPri)
      if (helicityOn) event.list(helicityOn);
    
    // Count FS charged hadrons, partons, and quarks
    int nc=0;
    int np=0;
    int nq=0;
    int nGam=0;
    int nStrange=0;
    int nCharm  =0;
    int nBottom =0;
    int nStrangeBorn = (event[6].idAbs() == 3) ? 2 : 0;
    int nCharmBorn   = (event[6].idAbs() == 4) ? 2 : 0;
    int nBottomBorn  = (event[6].idAbs() == 5) ? 2 : 0;
    vector<int> iStrange;
    vector<int> iCharm;
    vector<int> iBottom;
    for (int i=1;i<event.size();i++) {
      // Plot x distributions of first two emissions
      if (i == 9) {
        HxFirst.fill(log(2*event[i].pAbs()/eCM));
      }
      if (i == 12) {
        if (event[i].id() == 21)
          HxSecond.fill(log(2*event[i].pAbs()/eCM));
        else
          HxSecondQuark.fill(log(2*event[i].pAbs()/eCM));
      }
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].status() > 80) nc++;
      // Find last parton-level partons
      int iDau1 = event[i].daughter1();
      if (iDau1 == 0 || abs(event[iDau1].status()) > 80) {
        // Count up partons and quarks
        if (event[i].isQuark() || event[i].isGluon()) np++;
        if (event[i].id() == 22) nGam++;
        if (event[i].isQuark()) nq++;
        if (event[i].idAbs() == 3) {
          nStrange++;
          HxStrange.fill(log(2*event[i].pAbs()/eCM));
          for (int j=0; j<(int)iStrange.size(); ++j) {
            double m2qq = m2(event[i],event[iStrange[j]]);
            HmSS.fill(sqrt(m2qq));
          }
          iStrange.push_back(i);
        }
        if (event[i].idAbs() == 4) {
          nCharm++;
          HxCharm.fill(log(2*event[i].pAbs()/eCM));
          for (int j=0; j<(int)iCharm.size(); ++j) {
            double m2qq = m2(event[i],event[iCharm[j]]);
            HmCC.fill(sqrt(m2qq));
          }
          iCharm.push_back(i);
        }
        if (event[i].idAbs() == 5) {
          nBottom++;
          HxBottom.fill(log(2*event[i].pAbs()/eCM));
          for (int j=0; j<(int)iBottom.size(); ++j) {
            double m2qq = m2(event[i],event[iBottom[j]]);
            HmBB.fill(sqrt(m2qq));
          }
          iBottom.push_back(i);
        }
      }
      if (event[i].isHadron() && event[i].isFinal()) {
        int idAbs = abs(event[i].id());
        if (idAbs > 1000 && idAbs < 10000) nBaryonSum++;
      }
    }
    HnQuarks.fill(0.5*nq,evwt);
    HnPartons.fill(1.0*np,evwt);
    HnCharged.fill(1.0*nc,evwt);
    HnGamma.fill(1.0*nGam,evwt);
    nQsum += 0.5*nq * evwt;
    nPsum += np * evwt;
    nCsum += nc * evwt;
    nStrangeSum += (nStrange - nStrangeBorn) * evwt;
    nCharmSum   += (nCharm - nCharmBorn) * evwt;
    nBottomSum  += (nBottom - nBottomBorn) * evwt;
    nGammaSum   += nGam * evwt;
    
    // Histogram thrust
    Thr.analyze( event );
    Hthr.fill(1.0-Thr.thrust(),wthr*evwt);
    HthrTail.fill(1.0-Thr.thrust(),wthrCoarse*evwt);
    Hmaj.fill(Thr.tMajor(),wShape*evwt);
    Hmin.fill(Thr.tMinor(),wShape*evwt);
    Hobl.fill(Thr.oblateness(),wShape*evwt);
    
    // Histogram Linear Sphericity values
    if (np >= 2.0) {
      SphLin.analyze( event );
      double evC=3*(SphLin.eigenValue(1)*SphLin.eigenValue(2)
                    + SphLin.eigenValue(2)*SphLin.eigenValue(3)
                    + SphLin.eigenValue(3)*SphLin.eigenValue(1));
      double evD=27*SphLin.eigenValue(1)*SphLin.eigenValue(2)
      *SphLin.eigenValue(3);
      HistC.fill(evC,wShape*evwt);
      HistD.fill(evD,wShape*evwt);
    }
    
  }
  
  //************************************************************************
  
  // POST-RUN FINALIZATION
  // Normalization, Statistics, Output
  
  //Normalize histograms to effective number of positive-weight events.
  double normFac=1.0/sumWeights;
  Hthr      *= normFac;
  HthrTail  *= normFac;
  HistC     *= normFac;
  HistD     *= normFac;
  Hmaj      *= normFac;
  Hmin      *= normFac;
  Hobl      *= normFac;
  HnQuarks    *= normFac;
  HnPartons   *= normFac;
  HnCharged   *= normFac;
  HnGamma     *= normFac;
  HxStrange   *= normFac;
  HxCharm     *= normFac;
  HxBottom    *= normFac;
  HxFirst    *= normFac;
  HxSecond    *= normFac;
  
  // Save a few tables to files
  /*
  HxStrange.table("strangeVJ.dat");
  HxCharm.table("charmVJ.dat");
  HxBottom.table("bottomVJ.dat");
  */

  // Print out (and possibly save) Vincia internal histograms
  if (vinciaOn && verbose >= 2) {
    vincia.printHistos();
    if (verbose >= 3) vincia.writeHistos("vincia01");
  }
  
  // Print a few histograms
  cout<<HnPartons<<endl;
  if (nGammaSum > 0) cout<<HnGamma<<endl;
  cout<<HnCharged<<endl;
  cout<<Hthr<<endl;
  cout<<HistC<<endl;
  cout<<HistD<<endl;
  cout<<HxFirst<<endl;
  cout<<HxSecond<<endl;
  cout<<HxSecondQuark<<endl;    
  cout<<HxStrange<<endl;
  cout<<HxCharm<<endl;
  cout<<HxBottom<<endl;  
  cout<<HmSS<<endl;
  cout<<HmCC<<endl;
  cout<<HmBB<<endl;

  // Print out Vincia end-of-run information
  pythia.stat();
  if (vinciaOn) vincia.printInfo();
  
  cout<<endl;
  cout<<fixed;
  cout<< " <nGluonSplit> = "<<num2str(nQsum * normFac-1.0)<<endl;
  cout<< " <nG->SS>      = "<<num2str(nStrangeSum * normFac / 2.)<<endl;
  cout<< " <nG->CC>      = "<<num2str(nCharmSum * normFac / 2.)<<endl;
  cout<< " <nG->BB>      = "<<num2str(nBottomSum * normFac / 2.)<<endl;
  cout<< " <nPartons>    = "<<num2str(nPsum * normFac)<<endl;
  cout<< " <nPhotons>    = "<<num2str(nGammaSum * normFac)<<endl;
  cout<< " <nCharged>    = "<<num2str(nCsum * normFac)<<endl;  
  cout<< " <nBaryons>    = "<<num2str(nBaryonSum * normFac)<<endl;  
  cout<<endl;
  
  // Done.
  return 0;
}
