// vincia03-root.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to study light-flavor fragmentation 
// at LEP (compared to L3 udsc measurements), using VINCIA's ROOT-based 
// runtime display. 
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (C) 2019 Peter Skands

// PYTHIA and VINCIA include files and namespaces
#include "Pythia8/Vincia.h"
using namespace Pythia8;

// ROOT: include files for linking to ROOT
#include "LiveDisplays/LiveDisplays.h"

// Global on/off switches for optional additional plots:
// 1) pTin and pTout distributions
const bool OPTshowPTinout = true;
// 2) identified-particle spectra
const bool OPTshowParticleSpectra = true;
// 3) identified-particle parent composition
const bool OPTshowParticleComposition = false;
// 4) initialise from saved runs?
// const bool OPTinitialiseFromSaved = false;

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects. 
  Pythia pythia; 
  VinciaPlugin vincia(&pythia);

  // Read user settings from file
  pythia.readFile("vincia03.cmnd");
 
  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Some useful pointers
  Settings* settingsPtr         = vincia.settingsPtr;        
  ParticleData* particleDataPtr = vincia.particleDataPtr; 
  
  // Extract settings to be used in the main program.  
  double eCM              = settingsPtr->parm("Beams:eCM");
  int    nEvent           = settingsPtr->mode("Main:numberOfEvents");  

  // Check if hadronisation is on (used to make plot labels)
  bool   hadronLevel      = settingsPtr->flag("HadronLevel:all");

  // Vincia settings
  bool   vinciaOn         = settingsPtr->flag("Vincia");
  if (!vinciaOn) {
    settingsPtr->word("LiveDisplays:legendLabel","Pythia");
  } else {
    settingsPtr->word("LiveDisplays:legendLabel","Vincia 2.3");
  }
  
  // Data path directory, extract from VINCIADATA
  string VINCIA_SHARE     = vincia.VINCIA_SHARE;
  string VINCIA_VAL       = VINCIA_SHARE+"validation/";
  string VINCIA_DAT       = VINCIA_SHARE+"data/";
  
  //************************************************************************
  // Initialize VINCIA (automatically initializes PYTHIA) 
  //************************************************************************

  // Initialize
  if(!vincia.init()) { return EXIT_FAILURE; }

  //************************************************************************
  // Book Histograms
  //************************************************************************

  // Create a LiveDisplaysApp: initialises & handles graphics communications
  LiveDisplaysApp liveDisplaysApp(&pythia);

  // Read files
  vector<string> mcFileRead;
  mcFileRead.push_back(string(VINCIA_VAL+"vincia2/vincia03").c_str());
  mcFileRead.push_back(string(VINCIA_VAL+"pythia/vincia03").c_str());
 
  // Prefix of rescue files (to initialise from last run)
  string rescuePrefix = "vincia03-";
  
  // Next event at which LiveDisplaysApp will save displays/histograms
  int iSave = liveDisplaysApp.nNextSave();

  // Vincia uncertainty bands: number of variations expected 
  int nWeights = max(1,vincia.nWeights());

  // Histogram bin and range variables
  int nBins;
  double xLo, xHi;

  // Book alphaS histogram
  nBins = 28;
  xLo   = -0.3;
  xHi   = 2.5;
  TH1D histAlphaS("alphaS","#alpha_{s}^{FSR}",nBins,xLo,xHi);
  TH1D histAlphaStwo(histAlphaS);
  TH1D histAlphaShalf(histAlphaS);
  TH1D histAlphaS1(histAlphaS);
  TH1D histAlphaSpy("alphaSpy","#alpha_{s}",nBins,xLo,xHi);
  TH1D histAlphaScmw("alphaScmw","#alpha_{s}",nBins,xLo,xHi);
  TH1D histAlphaSmsb("alphaSmsb","#alpha_{s}",nBins,xLo,xHi);
  TH1D histAlphaSdef("alphaSdef","#alpha_{s}",nBins,xLo,xHi);
  vector<TH1D*> histVecAlphaS;
  histVecAlphaS.push_back(&histAlphaS);
  histVecAlphaS.push_back(&histAlphaShalf);
  histVecAlphaS.push_back(&histAlphaStwo);
  // Set properties 
  histAlphaS.SetMinimum(0.05);
  histAlphaS.SetMaximum(4.0);
  histAlphaSmsb.SetMinimum(0.05);
  histAlphaSmsb.SetMaximum(4.0);
  histAlphaS.GetXaxis()->SetTitle("Log10(p_{T}) [GeV]");
  histAlphaSmsb.GetXaxis()->SetTitle("Log10(p_{T}) [GeV]");
  histAlphaS.GetYaxis()->SetTitle("Value of #alpha_{s}");
  histAlphaSmsb.GetYaxis()->SetTitle("Value of #alpha_{s}");
  LiveDisplay dispAlphaS(&histAlphaSmsb, &pythia, 460, 320);
  dispAlphaS.normalize(false);
  dispAlphaS.logY(true);
  dispAlphaS.ratioRange(0.9,1.75);
  
  // Fill histogram
  if (vinciaOn) {
    // Current VINCIA
    bool useCMW            = settingsPtr->flag("Vincia:useCMW");
    AlphaStrong* alphaSptr = &vincia.vinCom.alphaStrong;
    if (useCMW) alphaSptr  = &vincia.vinCom.alphaStrongCMW;
    // Default VINCIA
    bool useCMWdef            = settingsPtr->flagDefault("Vincia:useCMW");
    AlphaStrong* alphaSdefPtr = &vincia.vinCom.alphaStrongDef;
    if (useCMWdef) alphaSdefPtr  = &vincia.vinCom.alphaStrongDefCMW;
    // Pythia Default (Monash tune)
    AlphaStrong alphaPyDef;
    alphaPyDef.init(0.1365, 1, 5, false);
    // Sherpa Default (2-loop CMW with aS(mZ)=0.1188)
    AlphaStrong alpha2Lcmw;
    alpha2Lcmw.init(0.1188, 2, 5, true);
    // Sherpa Default (2-loop CMW with aS(mZ)=0.1188)
    AlphaStrong alpha2Lmsb;
    alpha2Lmsb.init(0.1188, 2, 5, false);
    dispAlphaS.addTH1DPtr(&histAlphaSpy);
    dispAlphaS.addTH1DPtr(&histAlphaScmw);
    //dispAlphaS.addTH1DPtr(&histAlphaS);
    for ( int iBin=1; iBin<=nBins; ++iBin ) {
      double logQ = xLo + (iBin-0.5)*(xHi-xLo)/nBins;
      double pT   = pow(10.,logQ);
      double kMu2 = settingsPtr->parm("Vincia:renormMultFacEmitF");
      double kPT2      = kMu2*pow2(pT);
      double mu2freeze = pow2(settingsPtr->parm("Vincia:alphaSmuFreeze"));
      double as        = alphaSptr->alphaS(kPT2 + mu2freeze);
      double asTwo     = alphaSptr->alphaS(4*(kPT2+mu2freeze));
      double asHalf    = alphaSptr->alphaS((kPT2+mu2freeze)/4.);
      histAlphaS.SetBinContent(iBin,as);
      histAlphaStwo.SetBinContent(iBin,asTwo);
      histAlphaShalf.SetBinContent(iBin,asHalf);
      histAlphaS.SetBinError(iBin,0.);
      histAlphaStwo.SetBinError(iBin,0.);
      histAlphaShalf.SetBinError(iBin,0.);
      // VINCIA Default 
      double kMu2def   = settingsPtr->parmDefault("Vincia:renormMultFacEmitF");
      double kPT2def   = kMu2def*pow2(pT);
      double mu2freezeDef = 
        pow2(settingsPtr->parmDefault("Vincia:alphaSmuFreeze"));
      double asDef     = alphaSdefPtr->alphaS(kPT2def + mu2freezeDef);
      histAlphaSdef.SetBinContent(iBin,asDef);
      histAlphaSdef.SetBinError(iBin,0.);      
      // PYTHIA Default
      double asPy   = alphaPyDef.alphaS(pow2(pT));
      histAlphaSpy.SetBinContent(iBin,asPy);
      histAlphaSpy.SetBinError(iBin,0.);
      // SHERPA Default
      double as2Lcmw   = alpha2Lcmw.alphaS(pow2(pT));
      histAlphaScmw.SetBinContent(iBin,as2Lcmw);
      histAlphaScmw.SetBinError(iBin,0.);
      // MSbar
      double as2Lmsb   = alpha2Lmsb.alphaS(pow2(pT));
      histAlphaSmsb.SetBinContent(iBin,as2Lmsb);
      histAlphaSmsb.SetBinError(iBin,0.);
    }
    vector<string> legendForAlphaS;
    legendForAlphaS.push_back("MSbar 0.1188 2L #scale[0.85]{n^{F}_{max}=5}");
    legendForAlphaS.push_back("Pythia Monash 2013 (0.1365 1L #scale[0.85]{n^{F}_{max}=5})");
    legendForAlphaS.push_back("Sherpa (CMW 0.1188 2L #scale[0.85]{n^{F}_{max}=5})");
    legendForAlphaS.push_back("Vincia");
    dispAlphaS.setLegend(legendForAlphaS);
    dispAlphaS.update();
  }


  // weights (for uncertainty bands)
  vector<TH1D*> vWeights;
  for (int i=0; i<nWeights; ++i)
    vWeights.push_back(new TH1D(string("weights"+num2str(i,0)).c_str(),
      "Event Weights",50, -2.0, 4.0));
  vWeights[0]->SetMinimum(0.005);
  vWeights[0]->SetMaximum(50.0);
  vWeights[0]->GetXaxis()->SetTitle("log_{10}(w)");
  vWeights[0]->GetYaxis()->SetTitle("Rate");
  if (nWeights >= 1) {
    LiveDisplay* dispWeights = new LiveDisplay(vWeights[0],&pythia);
    for (int i=1; i<nWeights; ++i)
      dispWeights->addTH1DPtr(vWeights[i]);
    dispWeights->showRatio(false);
    vector<string> bandLabels;
    bandLabels.push_back(pythia.info.weightLabel(0));
    bandLabels.push_back(pythia.info.weightLabel(1));
    bandLabels.push_back(pythia.info.weightLabel(2));
    bandLabels.push_back(pythia.info.weightLabel(3));
    bandLabels.push_back(pythia.info.weightLabel(4));
    dispWeights->setLegend(bandLabels);    
  }
  
  // nPartons
  TH1D *histNpartons=new TH1D("nPartons","N_{Partons}(Q_{had})", 50, -0.5, 49.5);
  histNpartons->SetMinimum(0.0005);
  histNpartons->SetMaximum(0.5);
  LiveDisplay dispNpartons(histNpartons,&pythia);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    dispNpartons.addHisFile(mcFileRead[iRead]+"-nPartons.dat");
  dispNpartons.logY(true);

  // Quark Pairs
  nBins = 10;
  xLo   = -0.5;
  xHi   = 9.5;
  vector<TH1D*> histVecnQuarks;
  histVecnQuarks.push_back(new TH1D("nQuarks","N_{Quarks}(Q_{had})",nBins,xLo,xHi));
  histVecnQuarks[0]->GetXaxis()->SetTitle("N_{Q}");
  histVecnQuarks[0]->GetYaxis()->SetTitle("P(N_{Q})");
  histVecnQuarks[0]->SetMinimum(0.00005);
  histVecnQuarks[0]->SetMaximum(5.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecnQuarks.push_back(new TH1D(string("nQuarks"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  LiveDisplay dispNquarks(histVecnQuarks,&pythia);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    dispNquarks.addHisFile(mcFileRead[iRead]+"-nQuarks.dat");
  dispNquarks.logY(true);

  // Number of final-state emissions
  nBins = 51;
  xLo   = -0.5;
  xHi   = 50.5;
  vector<TH1D*> histVecNfsr;
  histVecNfsr.push_back(new TH1D("Nfsr","N_{FSR}(Q_{had})", nBins, xLo, xHi));
  histVecNfsr[0]->GetXaxis()->SetTitle("N_{FSR}");
  histVecNfsr[0]->GetYaxis()->SetTitle("P(N_{FSR})");
  histVecNfsr[0]->SetMinimum(0.0005);
  histVecNfsr[0]->SetMaximum(0.5);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecNfsr.push_back(new TH1D(string("Nfsr"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  LiveDisplay dispNFSR(histVecNfsr,&pythia);
  dispNFSR.logY(true);

  // Mesons
  nBins = 10;
  xLo   = 0.5;
  xHi   = xLo + nBins;
  vector<TH1D*> histVecMesons;
  histVecMesons.push_back(new TH1D("Mesons","Meson Fractions",nBins,xLo,xHi));
  histVecMesons[0]->GetYaxis()->SetTitle("#frac{<n>}{<n_{ch}>}");
  histVecMesons[0]->GetXaxis()->SetBinLabel(1,"#pi^{#pm}");
  histVecMesons[0]->GetXaxis()->SetBinLabel(2,"#pi^{0}");
  histVecMesons[0]->GetXaxis()->SetBinLabel(3,"K^{#pm}");
  histVecMesons[0]->GetXaxis()->SetBinLabel(4,"#eta");
  histVecMesons[0]->GetXaxis()->SetBinLabel(5,"#eta'");
  histVecMesons[0]->GetXaxis()->SetBinLabel(6,"#rho^{#pm}");
  histVecMesons[0]->GetXaxis()->SetBinLabel(7,"#rho^{0}"); 
  histVecMesons[0]->GetXaxis()->SetBinLabel(8,"#omega");
  histVecMesons[0]->GetXaxis()->SetBinLabel(9,"K*^{#pm}");
  histVecMesons[0]->GetXaxis()->SetBinLabel(10,"#phi");
  histVecMesons[0]->GetXaxis()->SetLabelOffset(0.04);
  histVecMesons[0]->GetXaxis()->SetLabelSize(0.07);
  histVecMesons[0]->SetMaximum(20.01); 
  histVecMesons[0]->SetMinimum(0.001); 
  if (hadronLevel) {
    LiveDisplay* dispMesons = new LiveDisplay(histVecMesons,&pythia);
    dispMesons->normalize(false);
    dispMesons->logY(true);
    //dispMesons.setRatioRange("Mesons",0.5,1.5);
    dispMesons->setDataFile(string(VINCIA_DAT+"LEP91-mesons.dat").c_str(),20,1,40,80);  
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispMesons->addHisFile(mcFileRead[iRead]+"-Mesons.dat");
  }

  // Baryons
  nBins = 11;
  xLo   = 0.5;
  xHi   = xLo + nBins;
  vector<TH1D*> histVecBaryons;
  histVecBaryons.push_back(new TH1D("Baryons","Baryon Fractions",nBins, xLo, xHi));  
  histVecBaryons[0]->GetYaxis()->SetTitle("#frac{<n>}{<n_{ch}>}");
  histVecBaryons[0]->GetXaxis()->SetBinLabel(1,"p^{#pm}");
  histVecBaryons[0]->GetXaxis()->SetBinLabel(2,"#Lambda");
  histVecBaryons[0]->GetXaxis()->SetBinLabel(3,"#Sigma^{#pm}");
  histVecBaryons[0]->GetXaxis()->SetBinLabel(4,"#Sigma^{0}");
  histVecBaryons[0]->GetXaxis()->SetBinLabel(5,"#Xi^{#pm}");
  histVecBaryons[0]->GetXaxis()->SetBinLabel(6,"#Delta^{++}");
  histVecBaryons[0]->GetXaxis()->SetBinLabel(7,"#Sigma^{#pm}*");
  histVecBaryons[0]->GetXaxis()->SetBinLabel(8,"#Xi^{0}*");
  histVecBaryons[0]->GetXaxis()->SetBinLabel(9,"#Omega");
  //histVecBaryons[0]->GetXaxis()->SetBinLabel(10,"#Lambda#bar{#Lambda}");
  //histVecBaryons[0]->GetXaxis()->SetBinLabel(11,"#Lambda#Lambda");
  histVecBaryons[0]->GetXaxis()->SetLabelOffset(0.04);
  histVecBaryons[0]->GetXaxis()->SetLabelSize(0.07);
  histVecBaryons[0]->SetMaximum(10.0); 
  histVecBaryons[0]->SetMinimum(0.00001); 
  if (hadronLevel) {
    LiveDisplay* dispBaryons = new LiveDisplay(histVecBaryons,&pythia);
    dispBaryons->normalize(false);
    dispBaryons->logY(true);
    //dispBaryons.setRatioRange("Baryons",0.5,1.5);
    dispBaryons->setDataFile(string(VINCIA_DAT+"LEP91-baryons.dat").c_str(),
      20,1,40,80);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispBaryons->addHisFile(mcFileRead[iRead]+"-Baryons.dat");
  }    

  // Ratios
  nBins = 12;
  xLo   = 0.5;
  xHi   = xLo + nBins;
  vector<TH1D*> histVecRatios;
  histVecRatios.push_back(new TH1D("Ratios","Particle Ratios",nBins, xLo, xHi));  
  histVecRatios[0]->GetYaxis()->SetTitle("Ratio");
  histVecRatios[0]->GetXaxis()->SetBinLabel(1,"K*/K");
  histVecRatios[0]->GetXaxis()->SetBinLabel(2,"#phi/K*");
  histVecRatios[0]->GetXaxis()->SetBinLabel(3,"#phi/K");
  histVecRatios[0]->GetXaxis()->SetBinLabel(4,"#phi/#rho");
  histVecRatios[0]->GetXaxis()->SetBinLabel(5,"#phi/#omega");
  histVecRatios[0]->GetXaxis()->SetBinLabel(6,"#phi/#pi^{#pm}");
  histVecRatios[0]->GetXaxis()->SetBinLabel(7,"#Lambda/K");
  histVecRatios[0]->GetXaxis()->SetBinLabel(8,"#Lambda/p");
  histVecRatios[0]->GetXaxis()->SetBinLabel(9,"#Sigma/p");
  histVecRatios[0]->GetXaxis()->SetBinLabel(10,"#Xi/#Lambda");
  histVecRatios[0]->GetXaxis()->SetBinLabel(11,"#Xi*/#Sigma*");
  histVecRatios[0]->GetXaxis()->SetBinLabel(12,"#Omega/#Xi*");
  histVecRatios[0]->GetXaxis()->SetLabelOffset(0.04);
  histVecRatios[0]->GetXaxis()->SetLabelSize(0.07);
  histVecRatios[0]->SetMaximum(50.0); 
  histVecRatios[0]->SetMinimum(0.001); 
  if (hadronLevel) {
    LiveDisplay* dispRatios = new LiveDisplay(histVecRatios,&pythia);
    dispRatios->normalize(false);
    dispRatios->logY(true);
    //dispRatios.setRatioRange(0.5,1.5);
    dispRatios->setDataFile(string(VINCIA_DAT+"LEP91-ratios.dat").c_str(),
      20,1,40,80);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispRatios->addHisFile(mcFileRead[iRead]+"-Ratios.dat");
  }    

  // Heavy Quarks
  nBins =  9;
  xLo   = -0.5;
  xHi   = 8.5;
  vector<TH1D*> histVecCharm;
  histVecCharm.push_back(new TH1D("Charm","Charm Rates",nBins, xLo, xHi));
  histVecCharm[0]->GetYaxis()->SetTitle("<n> or BR(Z#rightarrowX)");
  histVecCharm[0]->GetXaxis()->SetBinLabel(1,"D^{#pm}");
  histVecCharm[0]->GetXaxis()->SetBinLabel(2,"D^{0}");
  histVecCharm[0]->GetXaxis()->SetBinLabel(3,"D^{*#pm}");
  histVecCharm[0]->GetXaxis()->SetBinLabel(4,"D_{s}^{#pm}");
  histVecCharm[0]->GetXaxis()->SetBinLabel(5,"#Lambda_{c}^{+}X");
  histVecCharm[0]->GetXaxis()->SetBinLabel(6,"g#rightarrowc#bar{c}");
  histVecCharm[0]->GetXaxis()->SetBinLabel(7,"J/#psi");
  histVecCharm[0]->GetXaxis()->SetBinLabel(8,"#chi_{c1}");
  histVecCharm[0]->GetXaxis()->SetBinLabel(9,"#psi_{3685}");
  histVecCharm[0]->GetXaxis()->SetLabelOffset(0.04);
  histVecCharm[0]->GetXaxis()->SetLabelSize(0.064);
  histVecCharm[0]->SetMaximum(50.0); 
  histVecCharm[0]->SetMinimum(0.0001); 
  if (hadronLevel) {
    LiveDisplay* dispCharm = new LiveDisplay(histVecCharm,&pythia);
    dispCharm->normalize(false);
    dispCharm->ratioRange(0.0,1.99);
    dispCharm->logY(true);
    dispCharm->setDataFile(string(VINCIA_DAT+"LEP-91-charm.dat").c_str(),
      20,40,60,70);  
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispCharm->addHisFile(mcFileRead[iRead]+"-Charm.dat");
  }

  // Beauty
  nBins =  8;
  xLo   = -0.5;
  xHi   = 7.5;
  vector<TH1D*> histVecBeauty;
  histVecBeauty.push_back(new TH1D("Beauty","Beauty Rates",nBins, xLo, xHi));
  histVecBeauty[0]->GetYaxis()->SetTitle("<n> or BR(Z#rightarrowX)");
  histVecBeauty[0]->GetXaxis()->SetBinLabel(1,"B^{+}X");
  histVecBeauty[0]->GetXaxis()->SetBinLabel(2,"B^{#pm0}");
  histVecBeauty[0]->GetXaxis()->SetBinLabel(3,"B^{*}_{uds}");
  histVecBeauty[0]->GetXaxis()->SetBinLabel(4,"B_{s}^{0}X");
  histVecBeauty[0]->GetXaxis()->SetBinLabel(5,"B_{bqq}X");
  histVecBeauty[0]->GetXaxis()->SetBinLabel(6,"g#rightarrowb#bar{b}");
  histVecBeauty[0]->GetXaxis()->SetBinLabel(7,"4b");
  histVecBeauty[0]->GetXaxis()->SetBinLabel(8,"#Upsilon(#times10)");
  histVecBeauty[0]->GetXaxis()->SetLabelOffset(0.04);
  histVecBeauty[0]->GetXaxis()->SetLabelSize(0.064);
  histVecBeauty[0]->SetMaximum(50.0); 
  histVecBeauty[0]->SetMinimum(0.00001); 
  if (hadronLevel) {
    LiveDisplay* dispBeauty = new LiveDisplay(histVecBeauty,&pythia);
    dispBeauty->normalize(false);
    dispBeauty->logY(true);
    dispBeauty->ratioRange(0.0,1.99);
    dispBeauty->setDataFile(string(VINCIA_DAT+"LEP-91-beauty.dat").c_str(),
      20,40,60,70);  
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispBeauty->addHisFile(mcFileRead[iRead]+"-Beauty.dat");
  }

  // m(bb) - 2mb_0
  nBins =  40;
  xLo   =   0.;
  xHi   =   4.;
  vector<TH1D*> histVecMbb;
  histVecMbb.push_back(new TH1D("mbb","(m(g#rightarrowb#bar{b})/2mb",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecMbb[0],rescuePrefix+"Mbb.dat");
  histVecMbb[0]->GetXaxis()->SetTitle("m_{b#bar{b}}/(2m_{b})");
  histVecMbb[0]->GetYaxis()->SetTitle("Rate");
  histVecMbb[0]->SetMinimum(0.5);
  histVecMbb[0]->SetMaximum(nEvent/1000.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecMbb.push_back(new TH1D(string("Mbb"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  LiveDisplay dispMbb(histVecMbb,&pythia);
  dispMbb.zeroSkipping(true);
  dispMbb.normalize(false);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    dispMbb.addHisFile(mcFileRead[iRead]+"-Mbb.dat");
  dispMbb.logY(true);
  // Add line at 1.0
  double yMin=histVecMbb[0]->GetMinimum();
  double yMax=histVecMbb[0]->GetMaximum();
  TLine* linePtr = new TLine(1.,yMin,1.,yMax);
  linePtr->SetLineColor(17);
  linePtr->SetLineWidth(1.);
  linePtr->SetLineStyle(7);
  dispMbb.addTLine(linePtr);

  // nCharged (udsc)
  nBins = 61;
  xLo   = -0.5;
  xHi   = 60.5;
  vector<TH1D*> histVecnCharged;
  histVecnCharged.push_back(new TH1D("nCh","Charged Multiplicity (udsc)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecnCharged[0],rescuePrefix+"Nch.dat");
  histVecnCharged[0]->GetXaxis()->SetTitle("n_{ch}");
  histVecnCharged[0]->GetYaxis()->SetTitle("Probability(n_{ch})");
  histVecnCharged[0]->SetMinimum(0.000001);
  histVecnCharged[0]->SetMaximum(200.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecnCharged.push_back(new TH1D(string("nCh"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel) {
    LiveDisplay* dispNch = new LiveDisplay(histVecnCharged,&pythia);
    dispNch->zeroSkipping(true);
    dispNch->setDataFile(string(VINCIA_DAT+"L3-91-Nch-udsc.dat").c_str(),
      20,20,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispNch->addHisFile(mcFileRead[iRead]+"-Nch.dat");
    dispNch->logY(true);
  }

  // Ln(1/x) (udsc)
  nBins = 45;
  xLo   = -0.2;
  xHi   = 8.8;
  vector<TH1D*> histVeclnx;
  histVeclnx.push_back(new TH1D("LnXInv","Charged Momentum Fraction (udsc)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVeclnx[0],rescuePrefix+"Lnx.dat");
  double wLnx = nBins/(xHi-xLo);
  histVeclnx[0]->GetXaxis()->SetTitle("|Ln(x_{p})|");
  histVeclnx[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/d|Ln(x_{p})|");
  histVeclnx[0]->SetMinimum(0.0004);
  histVeclnx[0]->SetMaximum(50.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVeclnx.push_back(new TH1D(string("LnXInv"+num2str(iVar,0)).c_str(),
			  " ",nBins,xLo,xHi));
  if (hadronLevel) {
    LiveDisplay* dispLnX = new LiveDisplay(histVeclnx,&pythia);
    dispLnX->setDataFile(string(VINCIA_DAT+"L3-91-LogInvx-udsc.dat").c_str(),
      10,30,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispLnX->addHisFile(mcFileRead[iRead]+"-LnXInv.dat");
    dispLnX->logY(true);
    dispLnX->rebinLikeData(true);
  }

  // Particle composition vs Ln(1/x) (udsc)
  vector<TH1D*> histVecidLnxSum;
  vector<TH1D*> histVecidLnx;
  histVecidLnxSum.push_back(new TH1D("IdLnXInvSum1", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVecidLnxSum.push_back(new TH1D("IdLnXInvSum2", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVecidLnxSum.push_back(new TH1D("IdLnXInvSum3", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVecidLnxSum.push_back(new TH1D("IdLnXInvSum4", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVecidLnx.push_back(new TH1D("IdLnXInv1",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVecidLnx.push_back(new TH1D("IdLnXInv2",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVecidLnx.push_back(new TH1D("IdLnXInv3",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVecidLnx.push_back(new TH1D("IdLnXInv4",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVecidLnx[0]->GetXaxis()->SetTitle("|Ln(x_{p})|");
  histVecidLnx[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ID}/d|Ln(x_{p})|");
  histVecidLnx[0]->SetMinimum(0.0);
  histVecidLnx[0]->SetMaximum(1.25);
  if (hadronLevel && OPTshowParticleComposition) { 
    LiveDisplay* dispLnXID = new LiveDisplay(histVecidLnx[0],&pythia);
    for (int i=1; i<int(histVecidLnx.size()); ++i) 
      dispLnXID->addTH1DPtr(histVecidLnx[i]);
    dispLnXID->logY(false);
    dispLnXID->normalize(false);
    vector<string> legendID;
    legendID.push_back("#pi^{#pm}");
    legendID.push_back("K^{#pm}");
    legendID.push_back("p^{#pm}");
    legendID.push_back("Other");
    dispLnXID->setLegend(legendID);
  }

  // Particle composition vs Ln(1/x) (udsc)
  vector<TH1D*> histVeckChLnxSum;
  vector<TH1D*> histVeckChLnx;
  nBins = 35;
  xLo   = -0.2;
  xHi   = 6.8;
  histVeckChLnxSum.push_back(new TH1D("KchLnXInvSum0", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnxSum.push_back(new TH1D("KchLnXInvSum1", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnxSum.push_back(new TH1D("KchLnXInvSum2", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnxSum.push_back(new TH1D("KchLnXInvSum3", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnxSum.push_back(new TH1D("KchLnXInvSum4", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnxSum.push_back(new TH1D("KchLnXInvSum5", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnxSum.push_back(new TH1D("KchLnXInvSum6", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnx.push_back(new TH1D("KchLnXInv0",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnx.push_back(new TH1D("KchLnXInv1",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnx.push_back(new TH1D("KchLnXInv2",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnx.push_back(new TH1D("KchLnXInv3",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnx.push_back(new TH1D("KchLnXInv4",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnx.push_back(new TH1D("KchLnXInv5",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnx.push_back(new TH1D("KchLnXInv6",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  histVeckChLnx[0]->GetXaxis()->SetTitle("|Ln(x_{p})|");
  histVeckChLnx[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{K}/d|Ln(x_{p})|");
  if (hadronLevel && OPTshowParticleComposition) {
    LiveDisplay* dispLnXKch = new LiveDisplay(histVeckChLnx[0],&pythia);
    for (int i=1; i<int(histVeckChLnx.size()); ++i) 
      dispLnXKch->addTH1DPtr(histVeckChLnx[i]);
    dispLnXKch->logY(false);
    dispLnXKch->normalize(false);
    vector<string> legendID;
    legendID.push_back("Total");
    legendID.push_back("s quarks");
    legendID.push_back("D and B mesons");
    legendID.push_back("K*");
    legendID.push_back("u and d quarks");
    legendID.push_back("c and b quarks");
    legendID.push_back("Other");
    dispLnXKch->setLegend(legendID);
    dispLnXKch->ratioRange(0.0,1.0);
  }

  // x (all)
  nBins = 200;
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> histVecx;
  histVecx.push_back(new TH1D("X","Charged Momentum Fraction",nBins,xLo,xHi));
  double wx = nBins/(xHi-xLo);
  histVecx[0]->GetXaxis()->SetTitle("x_{p}");
  histVecx[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dx_{p}");
  histVecx[0]->SetMinimum(0.00005);
  histVecx[0]->SetMaximum(40.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecx.push_back(new TH1D(string("X"+num2str(iVar,0)).c_str(),
        " ",nBins,xLo,xHi));
  if (hadronLevel) {
    LiveDisplay* dispX = new LiveDisplay(histVecx,&pythia);
    dispX->setDataFile(string(VINCIA_DAT+"Aleph-91-x.dat").c_str(),
      10,30,40,50,60);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispX->addHisFile(mcFileRead[iRead]+"-x.dat");
    dispX->logY(true);
    dispX->rebinLikeData(true);
  }

  // Auxiliary histogram for storing sum(|p|)
  vector<TH1D*> histVecpSum;
  histVecpSum.push_back(new TH1D("pSum","|p| vs Charged Momentum Fraction",nBins,xLo,xHi));
  
  // Rapidity wrt Thrust
  nBins = 24;
  xLo   = 0.0;
  xHi   = 6.0;
  vector<TH1D*> histVecyThr;
  histVecyThr.push_back(new TH1D("yThr","Charged Rapidity wrt Thrust",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecyThr[0],rescuePrefix+"yThr.dat");
  double wyThr = nBins/(xHi-xLo);
  histVecyThr[0]->GetXaxis()->SetTitle("|y|");
  histVecyThr[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/d|y|");
  histVecyThr[0]->SetMinimum(0.00005);
  histVecyThr[0]->SetMaximum(60.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecyThr.push_back(new TH1D(string("yThr"+num2str(iVar,0)).c_str(),
			  " ",nBins,xLo,xHi));
  LiveDisplay dispYthr(histVecyThr,&pythia);
  dispYthr.logY(true);  
  dispYthr.setDataFile(string(VINCIA_DAT+"Aleph-91-yThr.dat").c_str(),
    10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    dispYthr.addHisFile(mcFileRead[iRead]+"-yThr.dat");

  
  // x(K+-) from ALEPH, used to zoom in on the soft part of the spectrum
  nBins = 210;  // (needed due to non-uniform bins in data)
  xLo   = -5.25;
  xHi   = 0.;
  vector<TH1D*> histVecxKch;
  histVecxKch.push_back(new TH1D("xKch","K^{#pm} momentum fraction (Combined)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecxKch[0],rescuePrefix+"xKch.dat");
  double wxKch = nBins/(xHi-xLo);
  histVecxKch[0]->GetXaxis()->SetTitle("ln(x_{p})");
  histVecxKch[0]->GetYaxis()->SetTitle("1/<n_{K}> dn_{K}/dln(x)");
  histVecxKch[0]->SetMinimum(0.0005);
  histVecxKch[0]->SetMaximum(60.);  
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecxKch.push_back(new TH1D(string("xKch"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    // Make this plot extra wide to show tiny bins near zero
    LiveDisplay* dispXKch = new LiveDisplay(histVecxKch,&pythia);
    dispXKch->setDataFile(string(VINCIA_DAT+"Aleph-Opal-Delphi-91-xK.dat").c_str(),13,33,43,50,60,71,72);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispXKch->addHisFile(mcFileRead[iRead]+"-xKch.dat");
    dispXKch->rebinLikeData(true);
    dispXKch->logY(true);
  }

  // x(phi) (heaviest meson spectrum)
  nBins = 200;  // (needed due to non-uniform bins in data)
  xLo   = 0.;
  xHi   = 1.;
  vector<TH1D*> histVecxPhi;
  histVecxPhi.push_back(new TH1D("xPhi","#phi momentum fraction",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecxPhi[0],rescuePrefix+"xPhi.dat");
  double wxPhi = nBins/(xHi-xLo);
  histVecxPhi[0]->GetXaxis()->SetTitle("x_{p}");
  histVecxPhi[0]->GetYaxis()->SetTitle("1/<n_{#phi}> dn_{#phi}/dx");
  histVecxPhi[0]->SetMinimum(0.01);
  histVecxPhi[0]->SetMaximum(20.);  
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecxPhi.push_back(new TH1D(string("xPhi"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    // Make this plot extra wide to show tiny bins near zero
    LiveDisplay* dispXphi = new LiveDisplay(histVecxPhi,&pythia);
    dispXphi->setDataFile(string(VINCIA_DAT+"Aleph-91-xphi.dat").c_str(),
                      10,20,30,40,80);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispXphi->addHisFile(mcFileRead[iRead]+"-xPhi.dat");
    dispXphi->ratioRange(0.5,1.99);
    dispXphi->rebinLikeData(true);
    dispXphi->logY(true);
  }
  
  // x(protons) ALEPH + SLD combined
  nBins = 253;  // (needed due to non-uniform bins in data)
  xLo   = -5;
  xHi   = 0.0;
  vector<TH1D*> histVecxp;
  histVecxp.push_back(new TH1D("xp","p+#bar{p} momentum fraction (Combined)",nBins,xLo,xHi));
  double wxp = nBins/(xHi-xLo);
  histVecxp[0]->GetXaxis()->SetTitle("x_{p}");
  histVecxp[0]->GetYaxis()->SetTitle("1/<n_{p}> dn_{p}/dx");
  histVecxp[0]->SetMinimum(0.0001);
  histVecxp[0]->SetMaximum(500.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecxp.push_back(new TH1D(string("xp"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    LiveDisplay* dispXproton = new LiveDisplay(histVecxp,&pythia);
    dispXproton->setDataFile(string(VINCIA_DAT+"ALEPH-SLD-91-xp.dat").c_str(),
      13,33,43,50,60,71,72);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispXproton->addHisFile(mcFileRead[iRead]+"-xp.dat");
    dispXproton->rebinLikeData(true);
    dispXproton->logY(true);
  }

  // Lnx(Lambda)
  nBins = 25;  // (needed due to non-uniform bins in data)
  xLo   = 0;
  xHi   = 5.0;
  vector<TH1D*> histVecxLam;
  histVecxLam.push_back(new TH1D("xLam","#Lambda^{0} momentum fraction (-ln[x])",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecxLam[0],rescuePrefix+"xLam.dat");
  double wxLam = nBins/(xHi-xLo);
  histVecxLam[0]->GetXaxis()->SetTitle("#xi_{p}");
  histVecxLam[0]->GetYaxis()->SetTitle("1/<n_{#Lambda}> dn_{#Lambda}/d#xi");
  histVecxLam[0]->SetMinimum(0.0);
  histVecxLam[0]->SetMaximum(0.7);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecxLam.push_back(new TH1D(string("xLam"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    LiveDisplay* dispXLambda = new LiveDisplay(histVecxLam,&pythia);
    dispXLambda->setDataFile(string(VINCIA_DAT+"Aleph-91-LogInvxL.dat").c_str(),
      10,30,40,50,60);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispXLambda->addHisFile(mcFileRead[iRead]+"-xLam.dat");
    dispXLambda->logY(false);
  }

  // x(Xi)
  nBins = 55;  // (needed due to non-uniform bins in data)
  xLo   = 0;
  xHi   = 0.55;
  vector<TH1D*> histVecxXi;
  histVecxXi.push_back(new TH1D("xXi","#Xi^{#pm} momentum fraction",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecxXi[0],rescuePrefix+"xXi.dat");
  double wxXi = nBins/(xHi-xLo);
  histVecxXi[0]->GetXaxis()->SetTitle("x_{E}");
  histVecxXi[0]->GetYaxis()->SetTitle("1/<n_{#Xi}> dn_{#Xi}/dx");
  histVecxXi[0]->SetMinimum(0.07);
  histVecxXi[0]->SetMaximum(50.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecxXi.push_back(new TH1D(string("xXi"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    LiveDisplay* dispXXi = new LiveDisplay(histVecxXi,&pythia);
    dispXXi->setDataFile(string(VINCIA_DAT+"Aleph-91-xXi.dat").c_str(),
      10,30,40,80,0);
    dispXXi->logY(true);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispXXi->addHisFile(mcFileRead[iRead]+"-xXi.dat");
    dispXXi->rebinLikeData(true);
  }

  // x(D*)
  nBins = 20;  
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> histVecxDstarCh;
  histVecxDstarCh.push_back(new TH1D("xDstarCh","D^{*#pm} energy fraction (x_{E}>0.1)",nBins,xLo,xHi));
  double wxDstarCh = nBins/(xHi-xLo);
  histVecxDstarCh[0]->GetXaxis()->SetTitle("x_{E}");
  histVecxDstarCh[0]->GetYaxis()->SetTitle("1/<n_{D}> dn_{D}/dx");
  histVecxDstarCh[0]->SetMinimum(0.0);
  histVecxDstarCh[0]->SetMaximum(2.8);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecxDstarCh.push_back(new TH1D(string("xDstarCh"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    LiveDisplay* dispXDstarCh = new LiveDisplay(histVecxDstarCh,&pythia);
    dispXDstarCh->setDataFile(string(VINCIA_DAT+"Aleph-91-xDstarCh.dat").c_str(),
      20,10,30,40,51,52,61,62);
    dispXDstarCh->logY(false);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispXDstarCh->addHisFile(mcFileRead[iRead]+"-xDstarCh.dat");
  }

  // x(D*) parents
  nBins = 20;  
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> histVecmotDstarCh;
  histVecmotDstarCh.push_back(new TH1D("motDstarCh","x(D^{*#pm}) (x_{E}>0.1)",
      nBins,xLo,xHi));
  double wmotDstarCh = nBins/(xHi-xLo);
  histVecmotDstarCh[0]->GetXaxis()->SetTitle("x_{E}");
  histVecmotDstarCh[0]->GetYaxis()->SetTitle("1/<n_{D}> dn_{D}/dx");
  for (int iMot=1; iMot<=3; iMot++) 
    histVecmotDstarCh.push_back(new TH1D(string("motDstarCh"+num2str(iMot,0)).c_str(),
                                   " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleComposition) {
    LiveDisplay* dispMotDstarCh = new LiveDisplay(histVecmotDstarCh[0],&pythia);
    for (int i=1; i<int(histVecmotDstarCh.size()); ++i) 
      dispMotDstarCh->addTH1DPtr(histVecmotDstarCh[i]);
    dispMotDstarCh->logY(false);
    dispMotDstarCh->normalize(false);
    vector<string> legendID;
    legendID.push_back("Total");
    legendID.push_back("Z#rightarrow cc");
    legendID.push_back("Z#rightarrow bb");
    legendID.push_back("Z#rightarrow uu+dd+ss");
    dispMotDstarCh->setLegend(legendID);
    dispMotDstarCh->ratioRange(0.0,1.0);
  }

  // PTIN vs Thrust Axes
  nBins = 140; // Important = multiplum of # of data bins, for ratio rebinning
  xLo =  0.0;
  xHi = 14.0;
  vector<TH1D*> histVecPTinThr;
  histVecPTinThr.push_back(new TH1D("PTinThr","Charged p_{T}^{in} (Thrust Axes)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecPTinThr[0],rescuePrefix+"PTinThr.dat");
  double wPTinThr = nBins/(xHi-xLo);
  histVecPTinThr[0]->GetXaxis()->SetTitle("p_{T}^{in}");
  histVecPTinThr[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  histVecPTinThr[0]->SetMinimum(0.000001);
  histVecPTinThr[0]->SetMaximum(11.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecPTinThr.push_back(new TH1D(string("PTinThr"+num2str(iVar,0)).c_str(),
        " ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    LiveDisplay* dispPTinThr = new LiveDisplay(histVecPTinThr,&pythia);
    dispPTinThr->setDataFile(string(VINCIA_DAT+"Delphi-91-PTIN.dat").c_str(),
      10,30,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispPTinThr->addHisFile(mcFileRead[iRead]+"-PTinThr.dat");
    dispPTinThr->logY(true);
    dispPTinThr->rebinLikeData(true);
  }

  // PTIN vs Sphericity Axes
  nBins = 140; // Important = multiplum of # of data bins, for ratio rebinning
  xLo =  0.0;
  xHi = 14.0;
  vector<TH1D*> histVecPTinSph;
  histVecPTinSph.push_back(new TH1D("PTinSph","Charged p_{T}^{in} (Sphericity Axes)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecPTinSph[0],rescuePrefix+"PTinSph.dat");
  double wPTinSph = nBins/(xHi-xLo);
  histVecPTinSph[0]->GetXaxis()->SetTitle("p_{T}^{in}");
  histVecPTinSph[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  histVecPTinSph[0]->SetMinimum(0.000001);
  histVecPTinSph[0]->SetMaximum(11.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecPTinSph.push_back(new TH1D(string("PTinSph"+num2str(iVar,0)).c_str(),
                                " ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    LiveDisplay* dispPTinSph = new LiveDisplay(histVecPTinSph,&pythia);
    dispPTinSph->setDataFile(string(VINCIA_DAT+"Aleph-91-PTIN-s.dat").c_str(),
      10,30,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispPTinSph->addHisFile(mcFileRead[iRead]+"-PTinSph.dat");
    dispPTinSph->logY(true); 
    dispPTinSph->rebinLikeData(true);
  }

  // PTOUT vs Thrust Axes
  nBins = 50; // Important = multiplum of # of data bins, for ratio rebinning
  xLo =  0.0;
  xHi =  5.0;
  vector<TH1D*> histVecPToutThr;
  histVecPToutThr.push_back(new TH1D("PToutThr","Charged p_{T}^{out} (Thrust Axes)", nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecPToutThr[0],rescuePrefix+"PToutThr.dat");
  double wPToutThr = nBins/(xHi-xLo);
  histVecPToutThr[0]->GetXaxis()->SetTitle("p_{T}^{out}");
  histVecPToutThr[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  histVecPToutThr[0]->SetMinimum(0.000001);
  histVecPToutThr[0]->SetMaximum(11.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecPToutThr.push_back(new TH1D(string("PToutThr"+num2str(iVar,0)).c_str()," ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    LiveDisplay* dispPToutThr = new LiveDisplay(histVecPToutThr,&pythia);
    dispPToutThr->setDataFile(string(VINCIA_DAT+"Delphi-91-PTOUT.dat").c_str(),
      10,30,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispPToutThr->addHisFile(mcFileRead[iRead]+"-PToutThr.dat");
    dispPToutThr->logY(true);
    dispPToutThr->rebinLikeData(true);
  }

  // PTOUT vs Sphericity Axes
  nBins = 50; // Important = multiplum of # of data bins, for ratio rebinning
  xLo =  0.0;
  xHi =  5.0;
  vector<TH1D*> histVecPToutSph;
  histVecPToutSph.push_back(new TH1D("PToutSph",
    "Charged p_{T}^{out} (Sphericity Axes)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecPToutSph[0],rescuePrefix+"PToutSph.dat");
  double wPToutSph = nBins/(xHi-xLo);
  histVecPToutSph[0]->GetXaxis()->SetTitle("p_{T}^{out}");
  histVecPToutSph[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  histVecPToutSph[0]->SetMinimum(0.000001);
  histVecPToutSph[0]->SetMaximum(11.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecPToutSph.push_back(new TH1D(string("PToutSph"+num2str(iVar,0)).c_str()," ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    LiveDisplay* dispPToutSph = new LiveDisplay(histVecPToutSph,&pythia);
    dispPToutSph->setDataFile(string(VINCIA_DAT+"Aleph-91-PTOUT-s.dat").c_str(),
      10,30,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispPToutSph->addHisFile(mcFileRead[iRead]+"-PToutSph.dat");
    dispPToutSph->logY(true);
    dispPToutSph->rebinLikeData(true);
  }
  
  // PT vs Linearised Sphericity Axes
  nBins = 20; 
  xLo =  0.0;
  xHi =  1.0;
  vector<TH1D*> histVecPTSph;
  histVecPTSph.push_back(new TH1D("PTSph",
    "Charged p_{T} (vs Linearised Ch+Neu Sphericity Axis)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecPTSph[0],rescuePrefix+"PTSph.dat");
  double wPTSph = nBins/(xHi-xLo);
  histVecPTSph[0]->GetXaxis()->SetTitle("p_{T}");
  histVecPTSph[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  histVecPTSph[0]->SetMinimum(0.0);
  histVecPTSph[0]->SetMaximum(2.0);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecPTSph.push_back(new TH1D(string("PTSph"+num2str(iVar,0)).c_str()," ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    LiveDisplay* dispPTSph = new LiveDisplay(histVecPTSph,&pythia);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispPTSph->addHisFile(mcFileRead[iRead]+"-PTSph.dat");
    dispPTSph->logY(false);
    dispPTSph->ratioRange(0.951,1.049);
  }

  // PT vs Linearised Sphericity Axes
  nBins = 20; 
  xLo =  0.0;
  xHi =  1.0;
  vector<TH1D*> histVecPTSph100;
  histVecPTSph100.push_back(new TH1D("PTSph100",
    "Charged p_{T} (with |p| > 0.1 GeV)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecPTSph100[0],rescuePrefix+"PTSph100.dat");
  double wPTSph100 = nBins/(xHi-xLo);
  histVecPTSph100[0]->GetXaxis()->SetTitle("p_{T}");
  histVecPTSph100[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  histVecPTSph100[0]->SetMinimum(0.0);
  histVecPTSph100[0]->SetMaximum(2.0);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecPTSph100.push_back(new TH1D(string("PTSph100"+num2str(iVar,0)).c_str()," ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    LiveDisplay* dispPTSph100 = new LiveDisplay(histVecPTSph100,&pythia);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispPTSph100->addHisFile(mcFileRead[iRead]+"-PTSph100.dat");
    dispPTSph100->logY(false);
    dispPTSph100->ratioRange(0.951,1.049);
  }

  // PT vs Linearised Sphericity Axes
  nBins = 20; 
  xLo =  0.0;
  xHi =  1.0;
  vector<TH1D*> histVecPTSph200;
  histVecPTSph200.push_back(new TH1D("PTSph200",
    "Charged p_{T} (with |p| > 0.2 GeV)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecPTSph200[0],rescuePrefix+"PTSph200.dat");
  double wPTSph200 = nBins/(xHi-xLo);
  histVecPTSph200[0]->GetXaxis()->SetTitle("p_{T}");
  histVecPTSph200[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  histVecPTSph200[0]->SetMinimum(0.0);
  histVecPTSph200[0]->SetMaximum(2.0);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecPTSph200.push_back(new TH1D(string("PTSph200"+num2str(iVar,0)).c_str()," ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    LiveDisplay* dispPTSph200 = new LiveDisplay(histVecPTSph200,&pythia);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      dispPTSph200->addHisFile(mcFileRead[iRead]+"-PTSph200.dat");
    dispPTSph200->logY(false);
    dispPTSph200->ratioRange(0.951,1.049);
  }

  
  // Thrust (non-B flavor)
  nBins = 20; // Important = multiplum of # of data bins, for ratio rebinning
  xLo = 0.0;
  xHi = 0.5;
  vector<TH1D*> histVecThrustLight;
  histVecThrustLight.push_back(new TH1D("ThrustLight","1-Thrust (udsc)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecThrustLight[0],rescuePrefix+"T.dat");
  double wThrust = nBins/(xHi-xLo);
  histVecThrustLight[0]->GetXaxis()->SetTitle("1-T (udsc)");
  histVecThrustLight[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/d(1-T)");
  histVecThrustLight[0]->SetMinimum(0.0001);
  histVecThrustLight[0]->SetMaximum(1000.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecThrustLight.push_back(new TH1D(string("ThrustLight"+num2str(iVar,0)).c_str()," ",nBins,xLo,xHi));  
  LiveDisplay dispT(histVecThrustLight,&pythia);
  dispT.setDataFile(string(VINCIA_DAT+"L3-91-Thrust-udsc.dat").c_str(),
    10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    dispT.addHisFile(mcFileRead[iRead]+"-ThrustLight.dat");
  dispT.logY(true);

  // C (light)
  nBins = 20; // Important = multiplum of # of data bins, for ratio rebinning
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> histVecCLight;
  histVecCLight.push_back(new TH1D("CLight","C Parameter (udsc)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecCLight[0],rescuePrefix+"C.dat");
  double wC = nBins/(xHi-xLo);
  histVecCLight[0]->GetXaxis()->SetTitle("C");
  histVecCLight[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dC");
  histVecCLight[0]->SetMinimum(0.0005);
  histVecCLight[0]->SetMaximum(200.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecCLight.push_back(new TH1D(string("CLight"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  LiveDisplay dispC(histVecCLight,&pythia);
  dispC.setDataFile(string(VINCIA_DAT+"L3-91-C-udsc.dat").c_str(),
    10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    dispC.addHisFile(mcFileRead[iRead]+"-CLight.dat");
  dispC.logY(true);

  // D (udsc)
  nBins = 25; // Important = multiplum of # of data bins, for ratio rebinning
  xLo   = 0.0;
  xHi   = 0.8;
  vector<TH1D*> histVecDLight;
  histVecDLight.push_back(new TH1D("DLight","D Parameter (udsc)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecDLight[0],rescuePrefix+"D.dat");
  double wD = nBins*1.0/(xHi-xLo);
  histVecDLight[0]->GetXaxis()->SetTitle("D");
  histVecDLight[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dD");
  histVecDLight[0]->SetMinimum(0.0005);
  histVecDLight[0]->SetMaximum(100.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecDLight.push_back(new TH1D(string("DLight"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  LiveDisplay dispD(histVecDLight,&pythia);
  dispD.setDataFile(string(VINCIA_DAT+"L3-91-D-udsc.dat").c_str(),
    10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    dispD.addHisFile(mcFileRead[iRead]+"-DLight.dat");
  dispD.logY(true);

  // Wide Jet Broadening
  nBins = 20; // Important = multiplum of # bins in data, for ratio rebinning
  xLo = 0.0;
  xHi = 0.3;
  vector<TH1D*> histVecBjWide;
  histVecBjWide.push_back(new TH1D("BWide","Wide Jet Broadening (udsc)",nBins,xLo,xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecBjWide[0],rescuePrefix+"BW.dat");
  double wBjWide = nBins/(xHi-xLo);
  histVecBjWide[0]->GetXaxis()->SetTitle("B_{W}");
  histVecBjWide[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dB_{W}");
  histVecBjWide[0]->SetMaximum(1000.0); 
  histVecBjWide[0]->SetMinimum(0.001);   
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecBjWide.push_back(new TH1D(string("BWide"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  LiveDisplay dispBW(histVecBjWide,&pythia);
  dispBW.setDataFile(string(VINCIA_DAT+"L3-91-BWide-udsc.dat").c_str(),
    20,10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    dispBW.addHisFile(mcFileRead[iRead]+"-BWide.dat");
  dispBW.logY(true);

  // Total Jet Broadening
  nBins = 20; // Must match or be multiplum of # bins in data, for ratio rebinning
  xLo = 0.0;
  xHi = 0.4;
  vector<TH1D*> histVecBjTot;
  histVecBjTot.push_back(new TH1D("BTot","Total Jet Broadening (udsc)",
                            nBins, xLo, xHi));
  //if (OPTinitialiseFromSaved)
  //  disp.initTH1Dptr(histVecBjTot[0],rescuePrefix+"BT.dat");
  double wBjTot = 200.0/0.4;
  histVecBjTot[0]->GetXaxis()->SetTitle("B_{T}");
  histVecBjTot[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dB_{T}");
  histVecBjTot[0]->SetMaximum(1000.0); 
  histVecBjTot[0]->SetMinimum(0.001);     
  for (int iVar=1; iVar<nWeights; iVar++) 
    histVecBjTot.push_back(new TH1D(string("BTot"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  LiveDisplay dispBT(histVecBjTot,&pythia);
  dispBT.setDataFile(string(VINCIA_DAT+"L3-91-BTot-udsc.dat").c_str(),
    20,10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    dispBT.addHisFile(mcFileRead[iRead]+"-BTot.dat");
  dispBT.logY(true);

  // Event analysis tools
  Thrust Thrust(1), ThrustCh(3);
  Sphericity SphLin(1,1),SphSqCh(2,3);;

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // Number of quarks, partons, and charged particles
  double weightSum    = 0.0;
  double nQuarksSum   = 0.0;
  double nPartonsSum  = 0.0;
  double nChargedSum  = 0.0;
  double nCharged2sum = 0.0;
  double nAllSumCentral     = 0.0;
  double nChargedSumCentral = 0.0;
  double nMesSum[20] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nBarSum[20] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nCharmSum[20] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0. ,0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nBeautySum[20] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0. ,0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nEvDch = 0.0;
  double nEvD0 = 0.0;
  double sumpt = 0.0;
  double sumpt2 = 0.0;
  double sumch = 0.0;
  
  // B quark Pythia mass
  double mbPythia = particleDataPtr->m0(5);
  
  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    // Generate next event
    if (!pythia.next()) continue;
    Event& event = pythia.event;

    // Histogram uncertainty weights
    for (int iVar=0; iVar<nWeights; iVar++) {
      vWeights[iVar]->Fill(log10(max(1.0e-9,vincia.weight(iVar))));
    }
      
    // Count FS charged hadrons, partons, and quarks
    int nCharged = 0;
    int nPartons = 0;
    int nQuarks  = 0;

    // Bools to accept/reject events with specific content
    bool isB       = false;
    bool isDch     = false;
    bool isD0      = false;
    bool hasLambdaC = false;
    bool hasB0s    = false;
    bool hasBplus  = false;
    bool is4b      = false;
    bool hasG2b    = false;
    bool hasG2c    = false;
    bool isBbaryon = false;
    vector<int> iBottomProd;
    for (int i=1;i<event.size();i++) {
      // Skip when we got to hadronization part
      if (event[i].isParton() && abs(event[i].status()) > 70)
        break;
      int id = event[i].id();
      int iMot1 = event[i].mother1();
      // int iMot2 = event[i].mother2();
      int idMot1 = (iMot1 != 0) ? event[iMot1].id() : 0;
      //int idMot2 = (iMot2 != 0) ? event[iMot2].id() : 0;
      if (abs(id) == 5) {
        isB = true;
        if (idMot1 == 21) hasG2b = true;
        if (idMot1 != id) iBottomProd.push_back(i);
      }
      else if (abs(id) == 4) {
        if (idMot1 == 21) hasG2c = true;
      }
    }
    if (iBottomProd.size() >= 4) is4b = true;
    // Loop over produced b quarks; plot g->bb invariant mass
    for (unsigned int i=0; i<iBottomProd.size(); ++i) {
      int i1 = iBottomProd[i];
      if (i1 <= 10) continue;
      for (unsigned int j=i+1; j<iBottomProd.size(); ++j) {
        int i2 = iBottomProd[j];
        if (i2 <= 10) continue;
        double m2bb = m2(event[i1].p(),event[i2].p());
        for (int iVar=0; iVar<nWeights; iVar++) 
          histVecMbb[0]->Fill(sqrt(m2bb)/(2*mbPythia),vincia.weight(iVar));
      }
    }

    // Main loop over event record
    for (int i=1;i<event.size();i++) {

      // Skip particles that are just copies
      if (event[i].daughter1() == event[i].daughter2() 
          && event[i].daughter1() != 0) continue;

      // Count up final-state charged hadrons
      double x  =  2.0*event[i].pAbs()/eCM;
      double xE =  2.0*event[i].e()/eCM;
      double xi = -log(x);
      int id    =  event[i].id();
      int idp   =  abs(id);
      if (event[i].isCharged() && event[i].status() > 80) {
	nCharged++;
        for (int iVar=0; iVar<nWeights; iVar++) {
          if (vincia.weight(iVar) > 0.0) {
            if (! isB) histVeclnx[iVar]->Fill(xi,wLnx*vincia.weight(iVar));
            if (x>0.02) {
	      histVecx[iVar]->Fill(x,wx*vincia.weight(iVar));
	    }
	  }
	}
        // Identified-particle composition as function of Lnx
        if (!isB) {
          if (idp == 211) histVecidLnxSum[0]->Fill( xi, wLnx * vincia.weight(0));
          else if (idp == 321) histVecidLnxSum[1]->Fill( xi, wLnx * vincia.weight(0));
          else if (idp == 2212) histVecidLnxSum[2]->Fill( xi, wLnx * vincia.weight(0));
          else histVecidLnxSum[3]->Fill( xi, wLnx * vincia.weight(0));
        }
      }
      // Find last parton-level partons
      int iDau1 = event[i].daughter1();
      if (iDau1 == 0 || abs(event[iDau1].status()) > 80) {
	// Count up partons and quarks
	if (event[i].isQuark() || event[i].isGluon() ) nPartons++;
	if (event[i].isQuark()) nQuarks++;	
      }
      // Find identified particles
      /* Mesons:
	# 211 0: e+ e- --> pi+- X
	# 111 1: e+ e- --> pi0 X
	# 321 2: e+ e- --> K+- X
	# 221 3: e+ e- --> eta X
	# 331 4: e+ e- --> etaprime(958) X
	# 213 5: e+ e- --> rho(770)+- X
	# 113 6: e+ e- --> rho(770)0 X
	# 223 7: e+ e- --> omega(782) X
	# 323 8: e+ e- --> K*(892)+- X
	# 333 9: e+ e- --> phi(1020) X
	#    10: K* /K- (ee from 10 GeV to 91 GeV + ALICE pp 1208.5717)
        #    11: phi/K* (ee from 10 GeV to 91 GeV + ALICE pp 1208.5717)
        #    12: phi/K- (ee from 10 GeV to 91 GeV + ALICE pp 1208.5717)
        #    13: phi/pi- (ee from 10 GeV to 91 GeV + ALICE pp 1208.5717)
	# 10111 10: e+ e- --> a0(980) X
	# 10221 11: e+ e- --> f0(980) X
      */
      /* Baryons:
	 # 2212      0 : e+ e- --> p (pbar) X
	 # 3122      1: e+ e- --> Lambda (Lambdabar) X
	 # 3122/2212 2: e+ e- --> Lambda (Lambdabar) / p(pbar)
	 # 3122/321  3: e+ e- --> Lambda (Lambdabar) / K+-
	 # 3222+3112 4: e+ e- --> Sigma+- (Sigmabar+-) X
	 # 3212      5: e+ e- --> Sigma0 (Sigmabar0) X
	 # 2224      6 : e+ e- --> Delta(1232)++ (Delta(1232)--) X
	 # 3224+3114 7: e+ e- --> Sigma(1385)+- (Sigmabar(1385)+-) X
	 # 3312      8: e+ e- --> Xi- (Xibar+) X
	 # 3324      9: e+ e- --> Xi(1530)0 (Xibar(1530)0) X
	 # 3334     10: e+ e- --> Omega- (Omegabar+) X 
      */
      /* Charm:
	 # 411    0: e+ e- --> D+- X (HEPDATA/ALEPH)
         # 421    1: e+ e- --> D0 (Dbar0) X (HEPDATA/ALEPH)
         # 413    2: e+ e- --> D*(2010)+- X (PDG S46)
 	 # 431    3: e+ e- --> Ds+- X (PDG S46)
         #        4: e+ e- --> Lambda_c+ X (PDG Z summary)
         #        5: g->cc (OPAL+ALEPH)
         # 443    6: e+ e- --> J/psi X (PDG)
         # 20443  7: e+ e- --> chi(c1) X (PDG)
         # 100443 8: e+ e- --> psi(3685) X (PDG)	
      */
      /* Beauty:
         # 511          0: BR(Z --> B+ X) (PDG Z summary)
         # 511+521      1: e+ e- --> B+-0 X (PDG S46)	
         # 513+523+533  2: e+ e- --> B*+-0 X (PDG S46)	
         # 531          3: BR(Z --> Bs0 X) (PDG Z summary)	
         # 5000 - 5999  4: BR(Z --> B-baryons X) (PDG Z summary)
         #              5: g -> bb
         #              6: BR(Z --> 4b X) (PDG Z summary)
         # 553 + 100553 7: e+e- --> Upsilon(1S+2S+3S) X (PDG Z summary)
      */
      if (idp == 211) nMesSum[1] += vincia.weight(0);
      if (idp == 111) nMesSum[2] += vincia.weight(0);
      if (idp == 321) {
	nMesSum[3] += vincia.weight(0);
        // SLD charged kaons
        /*
	if ( (x >= 0.016 && x <= 0.164) 
	     || (x >= 0.208 && x <= 0.768) ) {
	  for (int iVar=0; iVar<nWeights; iVar++) 
	    histVecxKchSLD[iVar]->Fill(x,wxKchSLD*vincia.weight(iVar)); 
	}
        */
        // ALEPH, DELPHI, OPAL charged kaons
	if ( (xi <= 5.2 && xi >= 4.0) || (xi <= 3.925) ) {
	  for (int iVar=0; iVar<nWeights; iVar++) {
	    histVecxKch[iVar]->Fill(-xi,wxKch*vincia.weight(iVar));         
          }
        }
        // Histogram kaon mothers
        int iMot = event[i].mother1();
        int idMot = event[iMot].id();
        histVeckChLnxSum[0]->Fill( xi, wLnx * vincia.weight(0));
        if (abs(idMot) >= 400 && abs(idMot) < 599) 
          histVeckChLnxSum[2]->Fill( xi, wLnx * vincia.weight(0));
        else if (abs(idMot) == 313 || abs(idMot) == 323) 
          histVeckChLnxSum[3]->Fill( xi, wLnx * vincia.weight(0));
        else if (abs(idMot) == 3) 
          histVeckChLnxSum[1]->Fill( xi, wLnx * vincia.weight(0));
        else if (abs(idMot) == 21 || abs(idMot) <= 2) 
          histVeckChLnxSum[4]->Fill( xi, wLnx * vincia.weight(0));
        else if (abs(idMot) <= 5)
          histVeckChLnxSum[5]->Fill( xi, wLnx * vincia.weight(0));
        else histVeckChLnxSum[6]->Fill( xi, wLnx * vincia.weight(0));
      }
      if (idp == 221) nMesSum[4] += vincia.weight(0);
      if (idp == 331) nMesSum[5] += vincia.weight(0);
      if (idp == 213) nMesSum[6] += vincia.weight(0);
      if (idp == 113) nMesSum[7] += vincia.weight(0);
      if (idp == 223) nMesSum[8] += vincia.weight(0);
      if (idp == 323) nMesSum[9] += vincia.weight(0);
      if (idp == 333) {
        nMesSum[10] += vincia.weight(0);
        if (x > 0.005) {
	  for (int iVar=0; iVar<nWeights; iVar++) {
	    histVecxPhi[iVar]->Fill(x,wxPhi*vincia.weight(iVar));         
          }
        }
      }
      if (idp == 10111) nMesSum[11] += vincia.weight(0);
      if (idp == 10221) nMesSum[12] += vincia.weight(0);
      if (idp == 2212) {
	nBarSum[1] += vincia.weight(0);
	// ALEPH + SLD combined protons
	if ( (xi <= 4.6047 && xi>=4.012 ) 
	     || (xi <= 3.735 && xi >= 3.578) 
	     || (xi <= 3.4189 && xi >= 1.799) 
	     || (xi <= 1.482) ) 
	  for (int iVar=0; iVar<nWeights; iVar++) {
	    histVecxp[iVar]->Fill(-xi,vincia.weight(iVar)*wxp);
          }
      }
      if (idp == 3122) {
	nBarSum[2] += vincia.weight(0);
	if (xi >= 0.4 && xi <= 4.8)
	  for (int iVar=0; iVar<nWeights; iVar++) {
	    histVecxLam[iVar]->Fill(xi,vincia.weight(iVar)*wxLam);
          }
      }
      if (idp == 3222) nBarSum[3] += vincia.weight(0); 
      if (idp == 3112) nBarSum[3] += vincia.weight(0); 
      if (idp == 3212) nBarSum[4] += vincia.weight(0); 
      if (idp == 3312) {
	nBarSum[5] += vincia.weight(0); 
	if (xE >= 0.03 && xE <= 0.5) 
	  for (int iVar=0; iVar<nWeights; iVar++) {
            // if (iVar != 0 && iVar != 9) continue;
	    histVecxXi[iVar]->Fill(xE,vincia.weight(iVar)*wxXi);
          }
      }
      if (idp == 2224) nBarSum[6] += vincia.weight(0);
      if (idp == 3114) nBarSum[7] += vincia.weight(0); 
      if (idp == 3224) nBarSum[7] += vincia.weight(0); 
      if (idp == 3324) nBarSum[8] += vincia.weight(0); 
      if (idp == 3334) nBarSum[9] += vincia.weight(0); 

      // Charm
      if (idp == 411) {
        isDch = true;
        nCharmSum[0] += vincia.weight(0);
      }
      if (idp == 421) {
        isD0 = true;
        nCharmSum[1] += vincia.weight(0);
      }
      // D*+-; include spectrum comparison
      if (idp == 413) {
        nCharmSum[2] += vincia.weight(0);
        for (int iVar=0; iVar<nWeights; iVar++) {
          if (xE > 0.1) {
            // if (iVar != 0 && iVar != 9) continue;
            histVecxDstarCh[iVar]->Fill(xE,vincia.weight(iVar)*wxDstarCh);
          }
        }        
        histVecmotDstarCh[0]->Fill(xE,wmotDstarCh*vincia.weight(0));
        if (abs(event[6].id()) == 4) 
          histVecmotDstarCh[1]->Fill(xE,wmotDstarCh*vincia.weight(0));
        else if (abs(event[6].id()) == 5) 
          histVecmotDstarCh[2]->Fill(xE,wmotDstarCh*vincia.weight(0));
        else 
          histVecmotDstarCh[3]->Fill(xE,wmotDstarCh*vincia.weight(0));
      }
      if (idp == 431) nCharmSum[3] += vincia.weight(0);
      if (idp == 443) nCharmSum[6] += vincia.weight(0);
      if (id == 4122) hasLambdaC = true;
      if (idp == 20443) nCharmSum[7] += vincia.weight(0);
      if (idp == 100443) nCharmSum[8] += vincia.weight(0);

      // Beauty
      if (id == 521) hasBplus = true;
      if (idp == 511 || idp == 521) nBeautySum[1] += vincia.weight(0);
      if (idp == 513 || idp == 523 || idp == 533)
        nBeautySum[2] += vincia.weight(0);
      if (id  == 531) hasB0s = true; 
      // (Upsilon rate is normalized by factor 10 to fit on plot)
      if (idp == 553 || idp == 100553) nBeautySum[7] += 10.*vincia.weight(0);
      if (idp >= 5000 && idp <= 6000) isBbaryon = true;
    }

    // Fill BR  inclusive branching fractions
    if (hasLambdaC) nCharmSum[4] += vincia.weight(0);
    if (hasG2c) nCharmSum[5] += vincia.weight(0);
    if (hasBplus) nBeautySum[0] += vincia.weight(0);
    if (hasB0s) nBeautySum[3] += vincia.weight(0);
    if (isBbaryon) nBeautySum[4] += vincia.weight(0);
    if (hasG2b) nBeautySum[5] += vincia.weight(0);
    if (is4b) nBeautySum[6] += vincia.weight(0);
    
    // Fill Histograms
    weightSum    += vincia.weight(0);
    nQuarksSum   += 0.5*nQuarks * vincia.weight(0);
    nPartonsSum  += nPartons * vincia.weight(0);
    nChargedSum  += nCharged * vincia.weight(0);
    nCharged2sum += pow2(nCharged) * vincia.weight(0); 
    nEvDch += (isDch ? 1 : 0) * vincia.weight(0);
    nEvD0 += (isD0 ? 1 : 0) * vincia.weight(0);
    histNpartons->Fill(1.0*nPartons,vincia.weight(0));    
    for (int iVar=0; iVar<nWeights; iVar++) {
      if (vincia.weight(iVar) <= 0.0) continue;
      if (! isB) histVecnCharged[iVar]->Fill(1.0*nCharged,vincia.weight(iVar));
      histVecnQuarks[iVar]->Fill(0.5*nQuarks,vincia.weight(iVar));
      if (vinciaOn)
	histVecNfsr[iVar]->Fill(vincia.fsrShower.getNbranch(1),vincia.weight(iVar));
      else 
	histVecNfsr[iVar]->Fill(pythia.info.nFSRinRes(),vincia.weight(iVar));
    }
    
    // Thrust
    Thrust.analyze( event );
    ThrustCh.analyze( event );
    // Force light quarks only
    if (! isB) {
      for (int iVar=0; iVar<nWeights; iVar++) {
        histVecThrustLight[iVar]->Fill(1-Thrust.thrust(),wThrust*vincia.weight(iVar));
      }
    }

    //TMP 
    // Histogram Sphericity values
    if (nCharged >= 0.0) {
      SphLin.analyze( event );
      SphSqCh.analyze( event );
      double evC=3*(SphLin.eigenValue(1)*SphLin.eigenValue(2)
		    + SphLin.eigenValue(2)*SphLin.eigenValue(3)
		    + SphLin.eigenValue(3)*SphLin.eigenValue(1));
      double evD=27*SphLin.eigenValue(1)*SphLin.eigenValue(2)
	*SphLin.eigenValue(3);
      // Light quarks
      if (! isB) {
        for (int iVar=0; iVar<nWeights; iVar++) {
          histVecCLight[iVar]->Fill(evC,wC*vincia.weight(iVar));
          histVecDLight[iVar]->Fill(evD,wD*vincia.weight(iVar));
        }
      }

      // Histogram jet masses
      Vec4 thrustAxis = Thrust.eventAxis(1);   
      Vec4 majorChAxis  = ThrustCh.eventAxis(2);   
      Vec4 minorChAxis  = ThrustCh.eventAxis(3);   
      Vec4 sphAxis1     = SphLin.eventAxis(1);
      Vec4 sphAxis2     = SphLin.eventAxis(2);
      Vec4 sphAxis3     = SphLin.eventAxis(3);
      //Vec4 sphChAxis    = SphSqCh.eventAxis(1);   
      Vec4 sphChAxis2   = SphSqCh.eventAxis(2);   
      Vec4 sphChAxis3   = SphSqCh.eventAxis(3);   
      Vec4 pPlus, pMinus;
      double bPlus(0.), bMinus(0.), pPlusAbs(0.), pMinusAbs(0.);
      double pToutThr(0.), pTinThr(0.), pToutSph(0.), pTinSph(0.);
      double pToutSphLin(0.), pTinSphLin(0.);
      // Loop over event to compute masses
      for (int i=0; i<event.size(); i++) {
        if (not event[i].isFinal() ) continue;
        int id = event[i].id();
        if (abs(id) == 12 || abs(id) == 14 || abs(id) == 16) continue;
        Particle& pNow = event[i];
        pTinSphLin  = abs(pNow.px()*sphAxis2.px() 
          + pNow.py()*sphAxis2.py() + pNow.pz()*sphAxis2.pz());
        pToutSphLin = abs(pNow.px()*sphAxis3.px() 
          + pNow.py()*sphAxis3.py() + pNow.pz()*sphAxis3.pz());
        double pTSphLin = sqrt(pow2(pTinSphLin) + pow2(pToutSphLin));
        sumpt += pTSphLin;
        sumpt2 += pow2(pTSphLin);
        sumch += 1.;
        // pTin and pTout
        if (pNow.isCharged() && OPTshowPTinout) {
          pTinThr  = abs(pNow.px()*majorChAxis.px() 
                     + pNow.py()*majorChAxis.py() + pNow.pz()*majorChAxis.pz());
          pToutThr = abs(pNow.px()*minorChAxis.px() 
                     + pNow.py()*minorChAxis.py() + pNow.pz()*minorChAxis.pz());
          pTinSph  = abs(pNow.px()*sphChAxis2.px() 
                     + pNow.py()*sphChAxis2.py() + pNow.pz()*sphChAxis2.pz());
          pToutSph = abs(pNow.px()*sphChAxis3.px() 
                     + pNow.py()*sphChAxis3.py() + pNow.pz()*sphChAxis3.pz());
          for (int iVar=0; iVar<nWeights; iVar++) {
            histVecPTinThr[iVar]->Fill(pTinThr,wPTinThr*vincia.weight(iVar));
            histVecPToutThr[iVar]->Fill(pToutThr,wPToutThr*vincia.weight(iVar));
            histVecPTinSph[iVar]->Fill(pTinSph,wPTinSph*vincia.weight(iVar));
            histVecPToutSph[iVar]->Fill(pToutSph,wPToutSph*vincia.weight(iVar));
            histVecPTSph[iVar]->Fill(pTSphLin,wPTSph*vincia.weight(iVar));
            if (pNow.pAbs() > 0.1)
              histVecPTSph100[iVar]->Fill(pTSphLin,
					  wPTSph100*vincia.weight(iVar));
            if (pNow.pAbs() > 0.2)
              histVecPTSph200[iVar]->Fill(pTSphLin,
					  wPTSph200*vincia.weight(iVar));
          }
        }
	// Also count density of all particles in central rapidity unit
	double pzSph =  pNow.px()*sphAxis1.px() 
	  + pNow.py()*sphAxis1.py() + pNow.pz()*sphAxis1.pz();
	double ySph = 0.5 * log( (pNow.e() + pzSph)/(pNow.e() - pzSph) );
	if (abs(ySph) < 0.5) {
	  nAllSumCentral += vincia.weight(0);
	  if (pNow.isCharged()) nChargedSumCentral += vincia.weight(0);
	}
        // Jet masses and broadening: sums
        double dotp = pNow.px()*thrustAxis.px()
          + pNow.py()*thrustAxis.py() + pNow.pz()*thrustAxis.pz();
        double pAbs = pNow.pAbs();
        double cosT = dotp/pAbs/thrustAxis.pAbs();
        double sinT = sqrt(1.-pow2(cosT));
        if (dotp > 0) {
          pPlus     += pNow.p();
          bPlus     += abs(pAbs * sinT);
          pPlusAbs  += pAbs;
        } else {
          pMinus    += pNow.p();
          bMinus    += abs(pAbs * sinT);
          pMinusAbs += pAbs;
        }
        // Charged-particle Rapidity wrt Thrust axis
        if (pNow.isCharged() || pNow.isGluon()) {
          double pz = pNow.px()*thrustAxis.px() 
            + pNow.py()*thrustAxis.py() + pNow.pz()*thrustAxis.pz();
          double E  = pNow.e();
          double yThr = (E-pz != 0.) ? 0.5*log( (E+pz) / (E-pz) ) : 10.0;
          for (int iVar=0; iVar<nWeights; iVar++) {
            histVecyThr[iVar]->Fill(abs(yThr),wyThr*vincia.weight(iVar));          
          }
        }
      }
      bPlus  /= 2.*(pPlusAbs+pMinusAbs);
      bMinus /= 2.*(pMinusAbs+pPlusAbs);    
      if (! isB) {
        for (int iVar=0; iVar<nWeights; iVar++) {
          histVecBjWide[iVar]->Fill(max(bPlus,bMinus),
                              wBjWide*vincia.weight(iVar));
          histVecBjTot[iVar]->Fill(bPlus+bMinus,wBjTot*vincia.weight(iVar));
        }
      }
    }

    // Finalise some histograms before updating displays
    if ( iEvent+1 >= liveDisplaysApp.nNextUpdate() ) {
      // Update Lnx histograms
      for (int iBin=0; iBin<int(histVeclnx[0]->GetNbinsX()); ++iBin) {
        double denom    = histVeclnx[0]->GetBinContent(iBin);
        double denomErr = histVeclnx[0]->GetBinError(iBin);
        for (int i=0; i<int(histVecidLnx.size()); ++i) {        
          double num    = histVecidLnxSum[i]->GetBinContent(iBin);
          double numErr = histVecidLnxSum[i]->GetBinError(iBin);
          if (num > 0. && denom >0.) {
            double val    = num/denom;
            double errRel = sqrt(pow2(denomErr/denom)+pow2(numErr/num));
            double err    = errRel * val;
            histVecidLnx[i]->SetBinContent(iBin,val);
            histVecidLnx[i]->SetBinError(iBin,err);
          } else {
            //VidLnx[i]->SetBinContent(iBin,0.5);
            //VidLnx[i]->SetBinError(iBin,0.5);
          }
        }
      }
      for (int iBin=0; iBin<int(histVeckChLnx[0]->GetNbinsX()); ++iBin) {
        for (int i=0; i<int(histVeckChLnx.size()); ++i) {        
          double num    = histVeckChLnxSum[i]->GetBinContent(iBin);
          double numErr = histVeckChLnxSum[i]->GetBinError(iBin);
          histVeckChLnx[i]->SetBinContent(iBin,num);
          histVeckChLnx[i]->SetBinError(iBin,numErr);
        }
      }
      // Update PID histograms
      if (nChargedSum > 0) {
        histVecMesons[0]->Reset();
        // Meson rates / <nCh>
        for (int iMes=1; iMes<=10; iMes++) {
	  // Normalise charged particles by extra factor 1/2 to get per-charge
	  double fac = 1.;
	  if (iMes == 1 || iMes == 3 || iMes == 6 || iMes == 9) fac = 0.5;
          histVecMesons[0]->SetBinContent(iMes,fac*nMesSum[iMes]/nChargedSum);
          double err = fac * sqrt(nMesSum[iMes]) / nChargedSum;
          histVecMesons[0]->SetBinError(iMes,err);
        }
        // Meson ratios
        histVecRatios[0]->SetBinContent(1,nMesSum[9]/nMesSum[3]);
        histVecRatios[0]->SetBinError(1,sqrt(nMesSum[9])/nMesSum[3]);
        histVecRatios[0]->SetBinContent(2,2.*nMesSum[10]/nMesSum[9]);
        histVecRatios[0]->SetBinError(2,2.*sqrt(nMesSum[10])/nMesSum[8]);
        histVecRatios[0]->SetBinContent(3,2.*nMesSum[10]/nMesSum[3]);
        histVecRatios[0]->SetBinError(3,2.*sqrt(nMesSum[10])/nMesSum[3]);
	double sumRho = nMesSum[6]+nMesSum[7];
        histVecRatios[0]->SetBinContent(4,3*nMesSum[10]/sumRho);
        histVecRatios[0]->SetBinError(4,3.*sqrt(nMesSum[10])/sumRho);
	histVecRatios[0]->SetBinContent(5,nMesSum[10]/nMesSum[8]);
	histVecRatios[0]->SetBinError(5,sqrt(nMesSum[10])/nMesSum[8]);
        histVecRatios[0]->SetBinContent(6,2.*nMesSum[10]/nMesSum[1]);
        histVecRatios[0]->SetBinError(6,2.*sqrt(nMesSum[10])/nMesSum[1]);
        // Baryon rates / <nCh>
	// Normalise by 1/2 to get per (anti)baryon
        for (int iBar=1; iBar<=9; iBar++) {
	  double fac = 0.5;
	  if (iBar == 10 || iBar == 11) fac = 1.0;
	  else if (iBar == 3 || iBar == 7) fac = 0.25;
          histVecBaryons[0]->SetBinContent(iBar,fac*nBarSum[iBar]/nChargedSum);
          double err = fac*sqrt(nBarSum[iBar]) / nChargedSum;
          histVecBaryons[0]->SetBinError(double(iBar)+1,err);
        }
        // Baryon ratios: fill Lambda/K, Lambda/p, etc
        histVecRatios[0]->SetBinContent(7,nBarSum[2]/nMesSum[3]);
        double err = sqrt(nBarSum[2]) / nMesSum[3];
        histVecRatios[0]->SetBinError(7,err);
        histVecRatios[0]->SetBinContent(8,nBarSum[2]/nBarSum[1]);
        err = sqrt(nBarSum[2]) / nBarSum[1];
        histVecRatios[0]->SetBinError(8,err);
        histVecRatios[0]->SetBinContent(9,(nBarSum[3]+nBarSum[4])/nBarSum[1]/3.);
        err = sqrt(nBarSum[3]+nBarSum[4]) / nBarSum[1];
        histVecRatios[0]->SetBinError(9,err);
        histVecRatios[0]->SetBinContent(10,nBarSum[5]/nBarSum[2]);
        err = sqrt(nBarSum[5]) / nBarSum[2];
        histVecRatios[0]->SetBinError(10,err);
        histVecRatios[0]->SetBinContent(11,2*nBarSum[8]/nBarSum[7]);
        err = sqrt(nBarSum[8]) / nBarSum[7];
        histVecRatios[0]->SetBinError(11,err);
        histVecRatios[0]->SetBinContent(12,nBarSum[9]/nBarSum[8]);
        err = sqrt(nBarSum[9]) / nBarSum[8];
        histVecRatios[0]->SetBinError(12,err);
        // Charm
        histVecCharm[0]->Reset();
        for (int iMes=0; iMes<=19; iMes++) {
          histVecCharm[0]->Fill(double(iMes),nCharmSum[iMes]/weightSum);
          err = sqrt(nCharmSum[iMes])/weightSum;
          histVecCharm[0]->SetBinError(double(iMes)+1,err);
        }
        // Beauty
        histVecBeauty[0]->Reset();
        for (int iMes=0; iMes<=19; iMes++) {
          histVecBeauty[0]->Fill(double(iMes),nBeautySum[iMes]/weightSum);
          err = sqrt(nBeautySum[iMes])/weightSum;
          // Note: Upsilon rate was artificially multiplied by 10
          if (iMes == 7) err *= sqrt(10.);
          histVecBeauty[0]->SetBinError(double(iMes)+1,err);
        }
      }
    }
    
    // Call LiveLiveDisplaysApp::update().
    // Write out extra info when hists were saved. 
    liveDisplaysApp.update(); 
    if ( iEvent >= iSave ) {
      iSave       = liveDisplaysApp.nNextSave();
      double rmsd = sqrt(nCharged2sum/weightSum-pow2(nChargedSum/weightSum));
      double rmsdn = rmsd/(nChargedSum/weightSum);
      cout<<" At event = "<<iEvent<<"   <nCh> = "<<nChargedSum/weightSum
	  <<" RMSD/<nCh> = "<<rmsdn
	  <<" <nPartons> = "<<nPartonsSum/weightSum
	  <<" <dNch/dy>|y=0 = "<<nAllSumCentral/weightSum
	  <<" ("<<nChargedSumCentral/weightSum<<")"
          <<" <pT> = "<<sumpt / sumch
          <<" <pT2> = "<<sumpt2 / sumch
	  <<"  (next save at "<<iSave<<")"<<endl;
      
    }
    
  }

  // End-of-run forced display update and saves
  liveDisplaysApp.update(true);
  
  // Print out Vincia internal histograms
  if (vinciaOn && settingsPtr->mode("Vincia:verbose") >= 2) vincia.printHistos();
  
  // Print out Vincia end-of-run information
  pythia.stat();
  if (vinciaOn) vincia.printInfo();

  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
